/*
 * Base64 encoding and decoding.
 * Copyright (C) 2001 Stephen Ostermiller <utils@Ostermiller.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * See COPYING.TXT for details.
 */


import java.io.*;


public class Base64{

    /**
     * This class need not be instantiated, all methods are static.
     */
    private Base64(){
    }

    protected static final byte[] base64Chars = {
        (byte)'A',(byte)'B',(byte)'C',(byte)'D',(byte)'E',(byte)'F',(byte)'G',(byte)'H',
        (byte)'I',(byte)'J',(byte)'K',(byte)'L',(byte)'M',(byte)'N',(byte)'O',(byte)'P',
        (byte)'Q',(byte)'R',(byte)'S',(byte)'T',(byte)'U',(byte)'V',(byte)'W',(byte)'X',
        (byte)'Y',(byte)'Z',(byte)'a',(byte)'b',(byte)'c',(byte)'d',(byte)'e',(byte)'f',
        (byte)'g',(byte)'h',(byte)'i',(byte)'j',(byte)'k',(byte)'l',(byte)'m',(byte)'n',
        (byte)'o',(byte)'p',(byte)'q',(byte)'r',(byte)'s',(byte)'t',(byte)'u',(byte)'v',
        (byte)'w',(byte)'x',(byte)'y',(byte)'z',(byte)'0',(byte)'1',(byte)'2',(byte)'3',
        (byte)'4',(byte)'5',(byte)'6',(byte)'7',(byte)'8',(byte)'9',(byte)'+',(byte)'/'
    };

    protected static final byte[] reverseBase64Chars = new byte[0xff];
    static {
        for (int i=0; i<reverseBase64Chars.length; i++){
            reverseBase64Chars[i] = -1;
        }
        for (byte i=0; i < base64Chars.length; i++){
            reverseBase64Chars[base64Chars[i]] = i;
        }
    }

    /**
     * Encodes and decodes each of the arguments.
     * Also check so see if encoding then decoding each
     * of the arguments returns the original argument.
     *
     * @param args Command line arguments to encode/decode
     */
    public static void main(String[] args){
        for (int i=0; i<args.length; i++){
            String encoded = encode(args[i]);
            String decoded = decode(encoded);
            System.out.println(args[i] + ": " + encoded + " " + decode(args[i]));
            if (!decoded.equals(args[i])){
                System.err.println("Encoding then decoding " + args[i] + " does not work. Decoded: '" + decoded + "'");
            }
        }
    }

    public static String encode(String string){
        return new String(encode(string.getBytes()));
    }


    public static byte[] encode(byte[] bytes){
        ByteArrayInputStream in = new ByteArrayInputStream(bytes);
        int mod;
        int length = bytes.length;
        if ((mod = length % 3) != 0){
            length += 3 - mod;
        }
        length = length * 4 / 3;
        ByteArrayOutputStream out = new ByteArrayOutputStream(length);
        try {
            encode(in, out, false);
        } catch (IOException x){
            // This can't happen.
            // The input and output streams were constructed
            // on memory structures that don't actually use IO.
        }
        return out.toByteArray();
    }

    public static void encode(InputStream in, OutputStream out, boolean lineBreaks) throws IOException {
        int[] inBuffer = new int[3];
        int lineCount = 0;

        boolean done = false;
        while (!done && (inBuffer[0] = in.read()) != -1){
            inBuffer[1] = in.read();
            inBuffer[2] = in.read();

            out.write(base64Chars[ inBuffer[0] >> 2 ]);
            if (inBuffer[1] != -1){
                out.write(base64Chars [(( inBuffer[0] << 4 ) & 0x30) | (inBuffer[1] >> 4) ]);
                if (inBuffer[2] != -1){
                    out.write(base64Chars [((inBuffer[1] << 2) & 0x3c) | (inBuffer[2] >> 6) ]);
                    out.write(base64Chars [inBuffer[2] & 0x3F]);
                } else {
                    out.write(base64Chars [((inBuffer[1] << 2) & 0x3c)]);
                    out.write('=');
                    done = true;
                }
            } else {
                out.write(base64Chars [(( inBuffer[0] << 4 ) & 0x30)]);
                out.write('=');
                out.write('=');
                done = true;
            }
            lineCount += 4;
            if (lineBreaks && lineCount >= 76){
                out.write('\n');
                lineCount = 0;
            }
        }
    }

    public static String decode(String string){
        return new String(decode(string.getBytes()));
    }

    public static String decode(String string, String enc) throws UnsupportedEncodingException {
        return new String(decode(string.getBytes(enc)), enc);
    }

    public static byte[] decode(byte[] bytes){
        ByteArrayInputStream in = new ByteArrayInputStream(bytes);
        int mod;
        int length = bytes.length;
        if ((mod = length % 4) != 0){
            length += 4 - mod;
        }
        length = length * 3 / 4;
        ByteArrayOutputStream out = new ByteArrayOutputStream(length);
        try {
            decode(in, out, false);
        } catch (IOException x){
            // This can't happen.
            // The input and output streams were constructed
            // on memory structures that don't actually use IO.
        }
        return out.toByteArray();
    }

    private static final int readBase64(InputStream in, boolean throwExceptions) throws IOException {
        int read;
        do {
            read = in.read();
            if (read == -1) return -1;
            if (throwExceptions && reverseBase64Chars[(byte)read] == -1 && 
                read != ' ' && read != '\n'  && read != '\r' && read != '\t' && read != '\f' && read != '='){
                throw new IOException ("Unexpected Base64 character: " + read);
            }
            read = reverseBase64Chars[(byte)read];
        } while (read == -1);
        return read;
    }

    public static void decode(InputStream in, OutputStream out, boolean throwExceptions) throws IOException {
        int[] inBuffer = new int[4];

        boolean done = false;
        while (!done && (inBuffer[0] = readBase64(in, throwExceptions)) != -1
            && (inBuffer[1] = readBase64(in, throwExceptions)) != -1){
            inBuffer[2] = readBase64(in, throwExceptions);
            inBuffer[3] = readBase64(in, throwExceptions);

            out.write(inBuffer[0] << 2 | inBuffer[1] >> 4);
            if (inBuffer[2] != -1){
                out.write(inBuffer[1] << 4 | inBuffer[2] >> 2);
                if (inBuffer[3] != -1){
                    out.write(inBuffer[2] << 6 | inBuffer[3]);
                } else {
                    done = true;
                }
            } else {
                done = true;
            }
        }
    }
}
